#!/usr/bin/perl

# Keep ssh agent pipes up to date so that I don't break the agent forwarding
# chain when I disconnect and reconnect a "screen" session

# This program is meant to be run from the .profile just before invoking
# screen. The first instance will run as a daemon. Later instances will
# notify the first instance and then quit.

# Warning: This makes use of /proc/${PID}/environ and /proc/${PID}/stat
# as implemented under Linux. If you wish to use this program on another
# OS, you will need to replace the code the which references those proc
# entries.

# Written 2008, William Herrin
# http://bill.herrin.us/freebies/

use strict;
use BSD::Resource;
use POSIX qw(setsid);
use User::pwent;
use IO::Select;
use Fcntl;
use FileHandle;

my $parent = getppid();
# background the program
if (my $pid = fork) {
  # Parent process returns
  exit (0);
}
# disconnect from the current sid/tty
setsid () or die "Can't start a new session: $!";
setpriority (0,0,10); # nice to 10

$SIG{HUP} = 'IGNORE';
my $pwent = getpwuid($<);
my $user = $pwent->[0];

my $dirname = $ENV{"HOME"}."/.ssh";
mkdir ($dirname);
chmod (0700, $dirname);
itsmineorfail ($dirname,0700);
my $pipe = "$dirname/agentd.pipe";

unless (lockagentd ($dirname)) {
  # I am not the first.
  itsmineorfail ($pipe,0600);
  open my $pipef, ">>$pipe";
  print $pipef $ENV{"SSH_AUTH_SOCK"} . "\n";
  close $pipef;
  rmscreenlogin ($parent);
  exit 0;
}

open my $log, ">>$dirname/agentd.log";
print $log time() . " sshagentd.pl starting.\n";
my $oldf = select $log; $|=1; select $oldf;

system "mknod $pipe p 2>>/dev/null";
chmod (0600,$pipe);
itsmineorfail ($pipe,0600);

my $agents = initagents();
my $agentlinks = {};

rmscreenlogin ($parent);

# Main select loop

sysopen my $pipef, $pipe, (O_RDONLY|O_NONBLOCK);
my $sel = new IO::Select ($pipef);
my ($pf,$piperead,@lines,$line,$inuse,$dead,$stillalive);
my $timeout = 5;
while (1) {
  ($pf) = $sel->can_read($timeout);
  if ($pf) {
    $sel->remove($pf);
    sysread $pf, $piperead, 2048;
    close ($pipef);
    sysopen my $pipef, $pipe, (O_RDONLY|O_NONBLOCK);
    $sel->add($pipef);
    @lines = split /\n/, $piperead;
    $timeout = 2;
    foreach $line (@lines) {
      next unless ($line =~ /\/agent.([0-9]+)$/);
      $agents->{$1} = $line;
      print $log time() . " pipe: $line\n";
    }
  }
  next unless ($dead = checkagents($agents));
  $inuse = usedagents($agents,$agentlinks);
  unless ($stillalive = liveagent($agents,$dead)) {
      print $log time() . " all agents dead.\n";
    unless ($inuse) {
      print $log time() . " no watched agents in use. Quitting.\n";
      close ($pipef);
      unlink $pipe;
      unlink "$dirname/agentd.lock";
      close ($log);
      exit 0;
    }
    # I still want an agent. Wait until I get one.
    $timeout = 86400;
    next;
  }
  $agentlinks = relinkdeadagents ($agents,$dead,
    $agents->{$stillalive},$agentlinks);
  $agentlinks = deleteunusedlinks ($agentlinks,$inuse,$agents->{$stillalive});
  fixdanglinglinks ($agentlinks,$agents->{$stillalive});
}

exit 0;

sub fixdanglinglinks {
  my ($links,$alive) = @_;

  foreach my $agent (sort keys %$links) {
    my $fn = $links->{$agent};
    next if ( -f $fn );
    relinkdeadagent($fn,$alive);
  }
  return 1;
}

sub deleteunusedlinks {
  my ($links,$inuse,$alive) = @_;

  foreach my $agent (sort keys %$links) {
    my $fn = $links->{$agent};
    next if $inuse->{$fn};
    deleteagentfiles($fn);
    delete ($links->{$agent});
  }
  return $links;
}

sub relinkdeadagents {
  my ($agents,$dead,$alive,$links) = @_;

  foreach my $agent (sort keys %$agents) {
    next unless ($dead->{$agent});
    print $log time() . " agent $agent is dead.\n";
    relinkdeadagent($agents->{$agent},$alive);
    $links->{$agent} = $agents->{$agent};
    delete ($agents->{$agent});
  }
  return $links; 
}

sub liveagent {
  my ($agents,$dead) = @_;

  foreach my $agent (sort keys %$agents) {
    return $agent unless ($dead->{$agent});
  }
  return undef;
}

sub itsmineorfail {
  # quit the program unless the named file belongs to my uid and has the
  # correct permissions.
  my ($fn,$wantmode) = @_;

  my ($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,$size,
    $atime,$mtime,$ctime,$blksize,$blocks) = stat ($fn);
  if ($uid != $<) {
    print STDERR "$fn owner $uid != $<\n";
    exit 1;
  }
  if (($mode & 07777) != $wantmode) {
    print STDERR "$fn owner mode " . sprintf ("%04o",$mode&07777) 
     . " != " . sprintf ("%04o",$wantmode) . "\n";
    exit 1;
  }
  return 0;
}

sub lockagentd {
  # attempt to become the primary sshagentd for this user. Return 1 on
  # success, 0 on failure.
  my ($dirname) = @_;

  my $lock = "$dirname/agentd.lock";
  open my $lockf, ">>$lock";
  print $lockf "$$\n";
  close $lockf;
  chmod (0600, $lock);
  itsmineorfail ($lock,0600);
  open $lockf, "<$lock";
  my $lockpid = <$lockf>;
  close $lockf;
  chop $lockpid;
  # If I got the lock, return 1.
  return 1 if ($lockpid == $$);
  # I didn't get the lock. Is the process who has the lock still running?
  open my $procf, "</proc/$lockpid/stat";
  my $line = <$procf>;
  close $procf;
  # If the process which has the lock is running. Return undef.
  return undef if ($line =~ /sshagentd.pl/);
  # The process which has the lock is defunct. Capture the stale lock.
  open my $lockf, ">$lock";
  print $lockf "$$\n";
  close $lockf;

  return 1;
}

sub initagents {
  # load currently defined ssh agents from /tmp
  my $agents = {};
  open my $agentlistf, "/bin/ls /tmp/ssh-*/agent.* 2>>/dev/null |";
  my $line;
  print $log time() . " Agents in use at start:\n";
  while ($line = <$agentlistf>) {
    next unless ($line =~ /agent.([0-9]+)$/);
    my $pid = $1;
    chop $line;
    $agents->{$pid} = $line;
    print $log time() . " $pid = $line\n";
  }
  close $agentlistf;
  return $agents;
}

sub checkagents {
  # return a hash of dead agents if any are dead
  my ($agents) = @_;

  my $dead = {};
  my $count = 0;
  foreach my $agent (sort keys %$agents) {
    open my $procf, "</proc/$agent/stat";
    my $line = <$procf>;
    close $procf;
    unless ($line =~ /sshd/) {
      $dead->{$agent} = $agents->{$agent};
      $count ++;
    }
  }
  return undef unless ($count>0);
  return $dead;
}

sub deleteagentfiles {
  # Remove the files associated with an ssh agent which is neither running
  # nor in use by any processes.
  my ($agent) = @_;

  unlink ($agent);
  my $agentdir = $agent;
  $agentdir =~ s/\/[^\/]*$//;
  print $log time() . " delete file=$agent, dir=$agentdir\n";
  return undef unless ($agentdir =~ /^\/tmp\//); # sanity check
  system "rm -rf $agentdir 2>>/dev/null";
  return $agentdir;
}

sub relinkdeadagent {
  my ($deadagent,$newagent) = @_;

  print $log time() . " relink: $deadagent -> $newagent\n";
  my $deaddir;
  return undef unless ($deaddir = deleteagentfiles($deadagent));
  my $newdir = $newagent;
  $newdir =~ s/\/[^\/]*$//;
  $newdir =~ s/^.*\///;
  print $log time() . " ln -s $newdir $deaddir\n";
  system "ln -s $newdir $deaddir";
  my $newagentf = $newagent;
  $newagentf =~ s/^.*\///;
  print $log time() . " ln -s $newagentf $deadagent\n";
  system "ln -s $newagentf $deadagent";
  return 1;
}

sub usedagents {
  # Determine which of the agents I'm watching are currently in use by
  # software on the system. Those which are not should be removed
  # from /tmp by some other function.
  my ($agenthash,@agents);

  print $log time() . " agents currently in use:\n";
  while ($agenthash = shift) {
    foreach my $pid (sort keys %$agenthash) {
      push @agents, $agenthash->{$pid};
      print $log time() . " agent=$pid," . $agenthash->{$pid} . "\n";;
    }
  }
  opendir my $dir, "/proc";
  my ($line,$pid);
  my $used = {}; 
  my $count = 0;
  while ($pid = readdir($dir)) {
    next unless ($pid =~ /^[0-9]+$/);
    next if ($pid == $$); # I don't need an agent so ignore me.
    undef $line;
    open my $env, "</proc/$pid/environ";
    read $env, $line, 20000;
    close $env;
    foreach my $agent (@agents) {
      if ($line =~ /SSH_AUTH_SOCK=$agent/) {
        $used->{$agent} = 1 ;
	      $count ++;
      }
    }
  }
  closedir ($dir);
  return undef unless ($count > 0);
  return $used;
}


sub rmscreenlogin {
  my ($parent) = @_;

  sleep (2);
  open my $f, "<" . $ENV{"HOME"}."/.screenlogin";
  my $pid = <$f>;
  close $f;
  chop $pid;
  return undef if ($pid != $parent);
  open $f, "</proc/$parent/stat";
  my $line = <$f>;
  close $f;
  return undef unless ($line =~ /bash/);
  unlink ($ENV{"HOME"}."/.screenlogin");
  return 1;
}

